#include "stdafx.h"
#include "PJNSNTP.h"
#include "SocMFC.h"
#include <cmath>


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


class CSNTPApp : public CWinApp
{
public:
//Methods
  BOOL InitInstance() override;

  DECLARE_MESSAGE_MAP()
};


class CMyCommandLineInfo : public CCommandLineInfo
{
public:
//Constructors / Destructors
  CMyCommandLineInfo() noexcept;
  CMyCommandLineInfo(const CMyCommandLineInfo&) = delete;
  CMyCommandLineInfo(CMyCommandLineInfo&&) = delete;
  ~CMyCommandLineInfo() = default; //NOLINT(modernize-use-override)

//Methods
  CMyCommandLineInfo& operator=(const CMyCommandLineInfo&) = delete;
  CMyCommandLineInfo& operator=(CMyCommandLineInfo&&) = delete;
  void ParseParam(const TCHAR* pszParam, BOOL bFlag, BOOL bLast) override;

//Member variables
  bool m_bSet;
};


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CSNTPApp, CWinApp) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

//The one and only MFC app instance
#pragma warning(suppress: 26426)
CSNTPApp theApp;


CMyCommandLineInfo::CMyCommandLineInfo() noexcept : m_bSet{false}
{
}

void CMyCommandLineInfo::ParseParam(const TCHAR* pszParam, BOOL bFlag, BOOL bLast)
{
  if (bFlag)
  {
    if (_tcsicmp(pszParam, _T("s")) == 0)
      m_bSet = true;
  }

  //Let the parent class do its stuff
  __super::ParseParam(pszParam, bFlag, bLast);
}


BOOL CSNTPApp::InitInstance()
{
  //Handle the command line
  CMyCommandLineInfo cmdInfo;
  ParseCommandLine(cmdInfo);

  //If there is not host specified, then log the error
  if (cmdInfo.m_strFileName.IsEmpty())
  {
    _tprintf(_T("testsntp host [/s]\n"));
    _tprintf(_T("where host is the NTP server address and\n"));
    _tprintf(_T("/s means synchronize the time instead of just displaying the retrieved values\n"));
    return FALSE;
  }

  //Initialise the winsock stack
  if (!AfxSocketInit())
  {
    _tprintf(_T("Failed to initialize Winsock, Error:%u\n"), GetLastError()); //NOLINT(clang-diagnostic-format)
    return FALSE;
  }

//#define TEST_SNTP_ROUNDTRIPPING
#ifdef TEST_SNTP_ROUNDTRIPPING
  //Test roundtripping of 1st January 1980 - 2050 through the CPJNNTPTime class
  for (WORD i{1980}; i<2050; i++)
  {
    SYSTEMTIME st1{};
    st1.wDay = 1;
    st1.wMonth = 1;
    st1.wYear = i;
    const CPJNNTPTime t{st1};
    SYSTEMTIME st2{t};
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wDay == st1.wDay);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wMonth == st1.wMonth);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wYear == st1.wYear);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wHour == st1.wHour);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wMinute == st1.wMinute);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wSecond == st1.wSecond);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wMilliseconds == st1.wMilliseconds);
  }

  //Test roundtripping of every second from 1st January 1980 - 1st January 2040 through the CPJNNTPTime class
  for (unsigned __int64 i{119600296960000000}; i<138534856960000000; i += 10000000)
  {
    FILETIME ft1{};
#pragma warning(suppress: 26472)
    ft1.dwLowDateTime = static_cast<DWORD>(i & 0x00000000FFFFFFFF);
#pragma warning(suppress: 26472)
    ft1.dwHighDateTime = static_cast<DWORD>((i & 0xFFFFFFFF00000000) >> 32);
    SYSTEMTIME st1{};
#pragma warning(suppress: 26496)
    AFXASSUME(FileTimeToSystemTime(&ft1, &st1));

    const CPJNNTPTime t{st1};
    const SYSTEMTIME st2{t};
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wDay == st1.wDay);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wMonth == st1.wMonth);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wYear == st1.wYear);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wHour == st1.wHour);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wMinute == st1.wMinute);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wSecond == st1.wSecond);
#pragma warning(suppress: 26496)
    AFXASSUME(st2.wMilliseconds == st1.wMilliseconds);
  }
#endif //#ifdef TEST_SNTP_ROUNDTRIPPING

  //Do the actual SNTP Query
  try
  {
    CPJNSNTPClient sntp;
    CPJNNTPServerResponse response;
    sntp.GetServerTime(cmdInfo.m_strFileName, response);

    _tprintf(_T("Time was successfully retrieved from NTP server\n"));

    const SYSTEMTIME st1{response.m_OriginateTime.AsSYSTEMTIME()};
    const SYSTEMTIME st2{response.m_ReceiveTime.AsSYSTEMTIME()};
    const SYSTEMTIME st3{response.m_TransmitTime.AsSYSTEMTIME()};
    const SYSTEMTIME st4{response.m_DestinationTime.AsSYSTEMTIME()};

    _tprintf(_T("                            DD/MM/YYYY  HH:MM:SS.MS\n"));
#pragma warning(suppress: 26472)
    _tprintf(_T("Client Originate Date was   %02d/%02d/%04d, %02d:%02d:%02d.%03d\n"), static_cast<int>(st1.wDay), static_cast<int>(st1.wMonth), static_cast<int>(st1.wYear), static_cast<int>(st1.wHour), static_cast<int>(st1.wMinute), static_cast<int>(st1.wSecond), static_cast<int>(st1.wMilliseconds));
#pragma warning(suppress: 26472)
    _tprintf(_T("Server Receive Date was     %02d/%02d/%04d, %02d:%02d:%02d.%03d\n"), static_cast<int>(st2.wDay), static_cast<int>(st2.wMonth), static_cast<int>(st2.wYear), static_cast<int>(st2.wHour), static_cast<int>(st2.wMinute), static_cast<int>(st2.wSecond), static_cast<int>(st2.wMilliseconds));
#pragma warning(suppress: 26472)
    _tprintf(_T("Server Transmit Date was    %02d/%02d/%04d, %02d:%02d:%02d.%03d\n"), static_cast<int>(st3.wDay), static_cast<int>(st3.wMonth), static_cast<int>(st3.wYear), static_cast<int>(st3.wHour), static_cast<int>(st3.wMinute), static_cast<int>(st3.wSecond), static_cast<int>(st3.wMilliseconds));
#pragma warning(suppress: 26472)
    _tprintf(_T("Client Destination Date was %02d/%02d/%04d, %02d:%02d:%02d.%03d\n"), static_cast<int>(st4.wDay), static_cast<int>(st4.wMonth), static_cast<int>(st4.wYear), static_cast<int>(st4.wHour), static_cast<int>(st4.wMinute), static_cast<int>(st4.wSecond), static_cast<int>(st4.wMilliseconds));
    _tprintf(_T("Round trip delay was %f seconds\n"), response.m_RoundTripDelay);
    _tprintf(_T("Local clock offset was %f seconds\n"), response.m_LocalClockOffset);
    _tprintf(_T("Response Stratum: %d\n"), response.m_nStratum);
    _tprintf(_T("Response Leap Indicator: %d\n"), response.m_nLeapIndicator);

    if (cmdInfo.m_bSet)
    {
      if (fabs(response.m_LocalClockOffset) < 3600)
      {
        //Get the current system time
        SYSTEMTIME st{};
        GetSystemTime(&st);
        FILETIME ft{};
        if (!SystemTimeToFileTime(&st, &ft))
        {
          _tprintf(_T("Failed to convert current SYSTEMTIME to FILETIME, Error:%u\n"), GetLastError()); //NOLINT(clang-diagnostic-format)
          return FALSE;
        }

        //Offset if by the local clock offset
#pragma warning(suppress: 26467 26472)
        const unsigned __int64 dwFT{(static_cast<unsigned __int64>(ft.dwHighDateTime) << 32) + static_cast<unsigned __int64>(ft.dwLowDateTime) + static_cast<unsigned __int64>(((response.m_LocalClockOffset * 10000000.0) + 0.5))};
#pragma warning(suppress: 26472)
        ft.dwLowDateTime = static_cast<DWORD>(dwFT & 0x00000000FFFFFFFF);
#pragma warning(suppress: 26472)
        ft.dwHighDateTime = static_cast<DWORD>((dwFT & 0xFFFFFFFF00000000) >> 32);
        if (!FileTimeToSystemTime(&ft, &st))
        {
          _tprintf(_T("Failed to convert new FILETIME to SYSTEMTIME, Error:%u\n"), GetLastError()); //NOLINT(clang-diagnostic-format)
          return FALSE;
        }

        //Set the time
        if (sntp.SetClientTime(st))
          _tprintf(_T("Time was successfully synchronized\n"));
        else
          _tprintf(_T("Failed to set the local time. Error:%u\n"), GetLastError()); //NOLINT(clang-diagnostic-format)
      }
      else
      {
        _tprintf(_T("Time difference was greater was 1 hour, not synchronizing clock\n"));
        _tprintf(_T("You should first set the time manually to the approximate time\n"));
      }
    } 
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
  catch(CWSocketException* pEx)
  {
    _tprintf(_T("Failed to retrieve time from NTP server, Error:%d\n"), pEx->m_nError);

    pEx->Delete();
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    _tprintf(_T("Failed to retrieve time from NTP server, Error:%d\n"), e.m_nError);
  }
#endif //#ifndef CWSOCKET_MFC_EXTENSIONS

  return FALSE;
}
